#include <math.h>
#include <stdio.h>
#include "/usr/local/express/include/express.h"

extern int NPROC;
#include "psa.h"
#include "decl.h"
#define DEBUG 0

/*-------------------------------------------------------------------------*/
float 
find_init_temp(vertex, max_vertices)
vertex_t *vertex;
int max_vertices;
{
	int local_index, send_proc;
	float tot_del_objfn, init_temp;

#if DEBUG
    fasync(stdout);
        printf("Into Find Init temp \n");
    fflush(stdout);
#endif

    local_index = rnd(0,max_vertices-1);

	while ( vertex[local_index].my_proc != mynode() ) 
		local_index = rnd(0,max_vertices-1);


    /* Select randomly a proc to send this vertex */

    send_proc = mynode();
    while (send_proc == mynode()) {
        if (mynode() == 0)
            send_proc = rnd((mynode()+1),NPROC-1);
        else 
            if (mynode() == NPROC-1)
                send_proc = rnd(0,(mynode()-1));
            else 
                send_proc = rnd(0,NPROC-1);
    }

#if DEBUG
	fasync(stdout);
	printf("\nProc:%d chose index:%d send_proc:%d\n",mynode(),local_index,
            send_proc);
	fflush(stdout);
#endif

	tot_del_objfn = find_tot_change_objfn(vertex,max_vertices,
                                          local_index, send_proc);
 
	init_temp = 20 * tot_del_objfn;

	/* Formula used : T0 = tot_del_objfn / ln (P)  where
	   P is the prob. equal to 0.95   */

	return init_temp;

}	
/*-------------------------------------------------------------------------*/
float
find_tot_change_objfn(vertex, max_vertices, i, send_proc)
vertex_t *vertex;
int max_vertices;
int i, send_proc;
{
	int j, cnt_loc_edges=0, load_source=0, load_dest=0; 
    int del_load_balance, type = 37;
	float cnt_comm=0.0, delta_objfn;

	for (j=0;j<max_vertices;j++)
		if (vertex[j].my_proc == mynode()) 
			load_source = load_source + vertex[j].deg;
		else if (vertex[j].my_proc == send_proc)
			load_dest = load_dest + vertex[j].deg;

	del_load_balance = 2*vertex[i].deg*( vertex[i].deg + load_dest 
			                             - load_source ); 



	for (j=0;j<vertex[i].deg;j++) {
		if (mynode() == vertex[i].proc[j]) {
			cnt_loc_edges--;
			cnt_comm = cnt_comm + vertex[i].egwt[j];
		}
		else 
			if (send_proc == vertex[i].proc[j])
				cnt_comm = cnt_comm - vertex[i].egwt[j];
	}

	cnt_loc_edges = cnt_loc_edges / 2;
	cnt_comm = cnt_comm / 2;

	delta_objfn = MU1*del_load_balance + MU2*cnt_comm;


	excombine( &delta_objfn, addf, sizeof(float), 1, ALLNODES, 
               NULLPTR, &type);

#if DEBUG
    fasync(stdout);
        printf("delta_objfn = %f \n",delta_objfn);
    fflush(stdout);
#endif

	
	return delta_objfn;
}                                  
/*-------------------------------------------------------------------------*/
float 
find_freez_temp()
{
	float t;
	t = TEMPFREEZ * (-1.0) / ((-31) * 0.6931 );
	/* Refers to -1/ ln 2^(-31) */
	return t;
}
/*--------------------------------------------------------------------------*/
int
find_vmvs(temp)
float temp;
{
	float vmvs;
	vmvs = MVS_CONST / temp;
	if (vmvs < 1)
		vmvs = 1;
	if (vmvs > 10)
		vmvs = 10;
	return vmvs;
}
/*--------------------------------------------------------------------------*/
int 
equilibrium(n_acc,n_att)
int n_acc, n_att;
{
	if ((n_acc < N_ACCMAX) || (n_att < N_ATTMAX)) 
		return 0;
	else
		return 1;
}
/*--------------------------------------------------------------------------*/
int
find_cut_edges( vertex, max_vertices)
vertex_t *vertex;
int max_vertices;
{
    int i, j, cut_edges = 0;

    for (i=0;i<max_vertices;i++) {
        for (j=0;j<vertex[i].deg;j++) {
            if (vertex[i].my_proc != vertex[i].proc[j])
                cut_edges++;
        }
    }
    cut_edges = cut_edges/2;
    return (cut_edges);
}
/*--------------------------------------------------------------------------*/
int
find_local_vertices( vertex, max_vertices)
vertex_t *vertex;
int max_vertices;
{
	int i, no_local_vertices=0;

	for (i=0;i<max_vertices;i++)
		if (vertex[i].my_proc == mynode())
			no_local_vertices++;

	return no_local_vertices;
}
/*-----------------------------------------------------------------------*/

void
generate_solution( vertex, max_vertices)
vertex_t *vertex;
int max_vertices;
{

    int i;
    FILE *fout;

	if (mynode() == 0) {
		fout = fopen("psa_out0", "wA");
		fasync(fout);
        for (i=0;i<max_vertices;i++)
            fprintf(fout,"GSN: %d PROC: %d\n",i, vertex[i].my_proc);

        fclose(fout);
    }
}
/*----------------------------------------------------------------------*/

