#include <stdio.h>
#include <stdlib.h>
#include <search.h>

#define MULTI_ON 1
#define DEBUG 0
#define TIME 1
#define INI_CUT 1
#define PRINT 0

#if MULTI_ON
#include "/usr/local/express/include/express.h"
#endif

#include "psa.h"
#include "decl.h"

int NPROC;

main( argc, argv)
int argc;
char **argv;
{

/*
c#######################################################################
c
c     Variable Declarations.
c
c#######################################################################
*/

   struct nodenv    env;	/* Express struct contains procnum etc */
   FILE            *fin;    /* Input File descriptor for Graph file */
   vertex_t      *vertex;   /* Vertex data Structure */
   int              *map;   /* Maps a local Index to Global Index */
   int      max_vertices;   /* No Of local vertices on mynode */
   float           temp0;   /* Initial temeratute for the System */
   float           tempF;   /* Freezing temerature for the System */
   float        my_objfn;   /* Contains initial objective function */
   int       converged=0;   /* Set to 1 when system converges */
   int  cnt_entry_sink=0;   /* No Of Entries in Sink data Struct */
   int     cnt_acc_mvs=0;   /* No Of Accepted Moves So far */
   int     cnt_att_mvs=0;   /* No Of attempted moves so far */
   int        accepted=0;   /* If a local perturbation is accepted then 1*/
   int              vmvs;   /* Frequecy of global updation */
   sink_t          *sink;   /* Ptr to sink data structure  */


/*
c#######################################################################
c
c     Misc. local variables:
c
c#######################################################################
*/

   int cnt_pass=0, no_cut_edges, no_local_vertices, tot_mvs=0;
   float temp, K, quality, lambda = 0.1;
   int glb_acc_mvs, glb_att_mvs;
   int type=99, max_load[32],maximum_load;
   int i, j, k, BLOCKSIZE, processor_no;
   long timeBegin, timeEnd;

/*
c@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
c
c     Beginning of executable code...
c
c@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

  /* Accept Command line arguments: File name for data graph */
   if ( argc != 5 ) 
	  exit(2);     /* Illegal Parameters on cmd line */
 
	sscanf(argv[2],"%f",&K);  
     
	sscanf(argv[3],"%f",&temp0); 

    sscanf(argv[4],"%d",&NPROC);

   /*  Read the Problem Graph to be Partitioned */

   init_express(&env);

   fin = open_file( argv[1], READ);


   max_vertices = read_max_vertices(fin);

#if DEBUG
   fasync(stdout);
   printf(" After Read_max_vertices %d\n", max_vertices);
   fflush(stdout);
#endif

   map = MEM(int, max_vertices);
   vertex = MEM(vertex_t, max_vertices);

   if ( (map == NULL) || ( vertex == NULL) )
	exit(10);   /* Out Of memory Condition */

   
   if ( read_graph(fin, vertex, max_vertices) < 0)
      exit(2);			/* Some error in Reading Graph */

#if DEBUG
   fasync(stdout);
   printf(" After Read_graph ..\n");
   fflush(stdout);
#endif

   fclose(fin);

/*-------------------------*/
/* Read Chow Wei's results */
/*-------------------------*/

    fin = open_file("color.4",READ);
    fscanf(fin,"%d",&max_vertices);
    for (i=0;i<max_vertices;i++) {
        fscanf(fin,"%d",&processor_no);
        vertex[i].my_proc = processor_no-1;
    }
    fclose(fin);

/*--------------------------*/
/* Reading done */
/*--------------------------*/

	/* Store info about procs to which nabors belong */
	for (i=0;i<max_vertices;i++) {
		vertex[i].proc = MEM(int, vertex[i].deg);
 		for (j=0;j<vertex[i].deg;j++) {
			vertex[i].proc[j] = vertex[ vertex[i].nabor[j] ].my_proc;
    	}
	}

#if INI_CUT 
    no_cut_edges = find_cut_edges( vertex, max_vertices);
	no_local_vertices = find_local_vertices( vertex, max_vertices);
	fasync(stdout);
	printf("Proc:: %d T0::%f, K:: %f, Initial cutset::%d, \
    Load :: %d\n", mynode(), temp0,K,no_cut_edges, 
    no_local_vertices);
	fflush(stdout);
#endif

/*---------------------------------------*/
/* Incremental Stuff starts here */
/*---------------------------------------*/
    tries = 0;
    cur_vertices = max_vertices;
	fincre = open_file("~/mesh/incre.1", READ);
    new_vertices = read_max_vertices(fincre);
	while((cur_vertices < max_vertices + new_vertices)
          &&( tries != MAX_TRY)) {
		for (i=max_vertices; i<= max_vertices+new_vertices-1;i++) {
			status = insert_vertex(i);
			if (status == 1) 
				cur_vertices++; 	
		}
		tries++;
	} /* end of while */
	fclose(fincre);

/*----------------------------------------*/
/* Incremental Stuff Ends here */
/*----------------------------------------*/

	/* Find Initial Objective Function and replicate in every
     * Processor
     */

	my_objfn = eval_init_objfn(vertex, max_vertices );

#if DEBUG
    fasync(stdout);
    	printf("Proc= %d Obj Func= %f\n", mynode(),  my_objfn);
	fflush(stdout);
#endif

	/*  Determine Initial Temperature T_0 */

#if DEBUG
    fasync(stdout);
        printf("Just Before Init temperature..\n");
    fflush(stdout);
#endif
/*	temp0 = find_init_temp(vertex, max_vertices);   */

#if DEBUG
    fasync(stdout);
        printf("Initial Temp found to be %f\n",temp0);
    fflush(stdout);
#endif

	/* Determine Freezing temperature */

	tempF = find_freez_temp();
#if DEBUG
    fasync(stdout);
        printf("After Freez temperature ..%f\n", tempF);
    fflush(stdout);
#endif

	/* Outter Loop for Simulated Annealing starts... */

	temp = temp0;     /* Start Annealing from temperature= temp0 */

	timeBegin = extime();

	while  (temp > tempF)  {
		vmvs = find_vmvs(temp);
#if DEBUG
    fasync(stdout);
        printf("Vmvs %d Temp %f\n", vmvs, temp);
    fflush(stdout);
#endif

		sink = MEM(sink_t,vmvs);
		if ( sink == NULL )
			exit(10);      /* Out Of memory Condition */
		cnt_acc_mvs = 0;
		glb_acc_mvs = 0;
		cnt_att_mvs = 0;
		glb_att_mvs = 0;

		while ( (! equilibrium(glb_acc_mvs, glb_att_mvs)) ||
                 (cnt_att_mvs % vmvs) )
        {
#if DEBUG
    fasync(stdout);
        printf("Inside Inner while..\n");
    fflush(stdout);
#endif
			accepted = local_perturb(vertex, max_vertices, 
                                     temp, sink, cnt_entry_sink );
			cnt_att_mvs++;

			if ( accepted ) {
				cnt_acc_mvs++;
				cnt_entry_sink++;
			}
#if DEBUG
    fasync(stdout);
if ( accepted ) {
        printf("Proc:%d, Petrurb Accepted, Pass:%d\n",mynode(),cnt_pass);
        printf("Proc:%d, cnt_acc_mvs:%d,cnt_entry_sink:%d\n",
                       mynode(),cnt_acc_mvs,cnt_entry_sink);
        for(i=0; i < cnt_entry_sink; i++) {
        printf("Proc:%d, sink[%d]: dest_proc:%d,lsn:%d\n",
                mynode(),i
               ,sink[i].dest_proc, sink[i].local_spin);
        }
    fflush(stdout);
}
#endif


			if ( ( cnt_att_mvs % vmvs ) == 0 ) {

				glb_synchronize( vertex, sink, max_vertices,
											  cnt_entry_sink, vmvs );

				cnt_entry_sink = 0;


	            my_objfn = eval_init_objfn(vertex, max_vertices );

                glb_att_mvs = cnt_att_mvs;
                glb_acc_mvs = cnt_acc_mvs;
                excombine( &glb_att_mvs, addi, sizeof(int),
                           1, ALLNODES, NULLPTR, &type);
                excombine( &glb_acc_mvs, addi, sizeof(int),
                           1, ALLNODES, NULLPTR, &type);

			}
#if PRINT
	fasync(stdout);
    no_cut_edges = find_cut_edges( vertex, max_vertices);
	no_local_vertices = find_local_vertices(vertex, max_vertices);
	printf("Proc: %d, Pass: %d, Objfn: %f, Cutset: %d ,Local vert: %d\n",
          mynode(),cnt_pass, my_objfn, no_cut_edges, no_local_vertices);
	fflush(stdout);
#endif


		} /* End of Inner loop which checks for Metropolis Criterion */
		
		free(sink);

        tot_mvs = tot_mvs + cnt_att_mvs;
 
#if DEBUG
    fasync(stdout);
        printf("Proc:%d, In Outa Loop\n", mynode());
    fflush(stdout);
#endif


		temp = K * temp;
		cnt_pass++;

	}  /* End Of Outter While loop for Simulated Annealing */

	timeEnd = extime();
#if TIME
	fasync(stdout);
	printf("\n Total Time to execute : %ld\n", timeEnd - timeBegin);
    printf("\nProc:: %d , Avg time per pass :: %ld\n",mynode(),
          (timeEnd - timeBegin)/tot_mvs);
	fflush(stdout);
#endif

#if INI_CUT
	no_cut_edges = find_cut_edges( vertex, max_vertices);
	no_local_vertices = find_local_vertices(vertex, max_vertices);
    max_load[mynode()] = no_local_vertices;
	fasync(stdout);
	printf("\nOuter Passes= %d, Total Passes= %d ,Cut_edges = %d, Local : %d\n"
           ,cnt_pass,tot_mvs,no_cut_edges, no_local_vertices);
    fflush(stdout);
           
/*    generate_solution( vertex, max_vertices); */
    maximum_load = 0;
    for (i=0;i<NPROC;i++)
		if (max_load[i] > maximum_load) 
 			maximum_load = max_load[i];
	quality =max_vertices/(NPROC*max_load[mynode()] + lambda*no_cut_edges);
    fasync(stdout);
    printf("\nProc :: %d ,Best Quality :: %f\n", mynode(),quality);
	fflush(stdout);

#endif

}  /* End Of Main Program */
/*-------------------------------------------------------------------*/
int 
init_express(pnode)
   struct nodenv  *pnode;
{

   exparam(pnode);
   return (pnode->procnum);	/* What's my proc_number */

}

/*-----------------------------------------------------------------*/
FILE *open_file(fname, mode)
char *fname;
int mode;
{

   FILE *fin;

	if (mode == READ)
		fin = fopen(fname, "rA");
	else {
		if (mode == WRITE) {
			sprintf(fname,"%s%d",fname,mynode());
			fin = fopen(fname, "wA");  /* Write to a different file */
		}
		else {
			fasync(stderr);
			fprintf(stderr, "\n Cannot open file: Illegal mode specified\n");
			fflush(stderr);
			exit(50);
		}
	}
   if ( fin == (FILE *) 0) {
      fasync(stderr);
      fprintf(stderr, "Node %d fails to open file\n", mynode());
      fflush(stderr);
      exit(13);
   }

   fasync(fin);
   return (fin);
}
/*-------------------------------------------------------------------------*/

