#include <stdio.h>
#include <search.h>

#include "/usr/local/express/include/express.h"
#include "psa.h"
#include "decl.h"

extern int NPROC;
#define DEBUG 0
#define PRINT 0

/*-------------------------------------------------------------------------*/

float eval_init_objfn(vertex, max_vertices )
vertex_t *vertex;
int max_vertices;
{
	int i, j, type = 37;           /* Type is arbit, express needs it! */
    int cnt_loc_edges = 0, cnt_degree=0;
    float cnt_comm = 0.0;
    float objfn_array[32], OBJFN = 0.0;
    /* 32 is the max no of procs we will use */


	for (i=0;i<max_vertices;i++) {
		if (vertex[i].my_proc == mynode()) {
			cnt_degree = cnt_degree + vertex[i].deg;

			for( j=0; j < vertex[i].deg; j++) {
				if ( mynode() == vertex[i].proc[j] )
					++cnt_loc_edges;
				else
					cnt_comm = cnt_comm + vertex[i].egwt[j];
			} 				
		}
	}


    /* New Objective Function Calculation */
	cnt_loc_edges = cnt_loc_edges / 2;
	cnt_comm = cnt_comm / 2;

    for( i=0; i< NPROC; i++)
		objfn_array[i] = 0.0;    /* Init the objfn array for express */

    objfn_array[mynode()] = MU1*cnt_degree*cnt_degree  + MU2*cnt_comm;

    excombine( objfn_array, addf, sizeof(float),
               NPROC, ALLNODES, NULLPTR, &type );

    
    for( i=0; i< NPROC; i++)
		OBJFN = OBJFN + objfn_array[i] ;

#if PRINT
    fasync(stdout);
        printf("LOCALLY Proc= %d Load Bal term: %d ,Comm=%f  Total ObjFn %f\n", 
                mynode(), cnt_degree*cnt_degree,cnt_comm, OBJFN);
    fflush(stdout);
#endif

    return OBJFN;
}
/*-------------------------------------------------------------------------*/
int maxint(p1, p2, size)
int *p1;
int *p2;
int size;
{

    if ( *p2 > *p1)
        *p1 = *p2;

    return 0;
}

/*------------------------------------------------------------------------*/
float
eval_del_objfn(vertex,max_vertices, i,sink,cnt_entry_sink, send_proc )
vertex_t *vertex;
int max_vertices;
int i;
sink_t *sink;
int cnt_entry_sink;
int send_proc;
{

	int j, cnt_loc_edges=0, load_source=0, load_dest=0; 
    int del_load_balance, SINKSIZE;
    float cnt_comm =0.0, delta_objfn;
	sink_t node, *node_ptr;

	for (j=0;j<max_vertices;j++)
		if (vertex[j].my_proc == mynode())
			load_source = load_source + vertex[j].deg;
		else if (vertex[j].my_proc == send_proc)
			load_dest = load_dest + vertex[j].deg;

    del_load_balance = 2*vertex[i].deg*( vertex[i].deg + load_dest
                                         - load_source );


	SINKSIZE = cnt_entry_sink;
	for (j=0;j<vertex[i].deg;j++) {
		if (mynode() == vertex[i].proc[j]) {
			/* Check to see if it is not already put in sink */
	   		node.local_spin = vertex[i].nabor[j];

			node_ptr = (sink_t *)bsearch((char *)(&node), (char *)sink,
				SINKSIZE, sizeof(sink_t),node_compare);

			if (node_ptr != NULL) {  /* Element present */
				cnt_comm = cnt_comm - vertex[i].egwt[j]; 	
			}
			else { /* Element not put in sink earlier */
				cnt_loc_edges--;
				cnt_comm = cnt_comm + vertex[i].egwt[j];
			}
		}
		else   /* Nabor is not on my proc */
			cnt_comm = cnt_comm - vertex[i].egwt[j];
	}
	delta_objfn = MU1*del_load_balance + MU2*cnt_comm;
	return (delta_objfn);
}
/*------------------------------------------------------------------------*/
int
node_compare(node1, node2)
sink_t *node1, *node2;
{
	return (node1->local_spin - node2->local_spin);
}
/*------------------------------------------------------------------------*/
/* Express Excombine Call wants these two functions: addf, addi */
/*------------------------------------------------------------------------*/

int
addf(p1, p2, size)
  float          *p1, *p2;
  int             size;
{
  *p1 += *p2;
  return (0);
}
/*------------------------------------------------------------------------*/

int
addi(p1, p2, size)
  int            *p1, *p2;
  int             size;
{
  *p1 += *p2;
  return (0);
}
/*------------------------------------------------------------------------*/

